import typing
from collections.abc import Callable

from loguru import logger
from notte_browser.session import NotteSession
from notte_browser.window import BrowserWindow
from notte_core.agent_types import AgentStepResponse
from notte_core.common.config import NotteConfig
from notte_core.credentials.base import BaseVault
from notte_core.storage import BaseStorage
from notte_sdk.types import AgentCreateRequest, AgentCreateRequestDict
from pydantic import field_validator

from notte_agent.agent import NotteAgent
from notte_agent.falco.perception import FalcoPerception
from notte_agent.falco.prompt import FalcoPrompt


class FalcoConfig(NotteConfig):
    enable_perception: bool = False

    @classmethod
    @field_validator("enable_perception")
    def check_perception(cls, value: bool) -> bool:
        if value:
            logger.warning("Perception should be disabled for falco. Don't set this argument to `True`.")
        return False


class FalcoAgent(NotteAgent):
    def __init__(
        self,
        window: BrowserWindow,
        storage: BaseStorage | None = None,
        vault: BaseVault | None = None,
        step_callback: Callable[[AgentStepResponse], None] | None = None,
        **data: typing.Unpack[AgentCreateRequestDict],
    ):
        _ = AgentCreateRequest.model_validate(data)
        config: FalcoConfig = FalcoConfig.from_toml(**data)
        session = NotteSession(window=window, storage=storage, enable_perception=False)
        super().__init__(
            prompt=FalcoPrompt(),
            perception=FalcoPerception(),
            config=config,
            session=session,
            vault=vault,
            step_callback=step_callback,
        )



from typing import final

from notte_browser.session import NotteSession, SessionTrajectoryStep
from notte_core.actions import BaseAction
from notte_core.browser.observation import StepResult
from notte_core.common.config import config
from notte_core.errors.base import NotteBaseError


class StepExecutionFailure(NotteBaseError):
    def __init__(self, message: str):
        super().__init__(
            user_message=message,
            agent_message=message,
            dev_message=message,
        )


class MaxConsecutiveFailuresError(NotteBaseError):
    def __init__(self, max_failures: int):
        self.max_failures: int = max_failures
        message = f"Max consecutive failures reached in a single step: {max_failures}."
        super().__init__(
            user_message=message,
            agent_message=message,
            dev_message=message,
        )


@final
class SafeActionExecutor:
    def __init__(
        self,
        session: NotteSession,
        max_consecutive_failures: int = config.max_consecutive_failures,
    ) -> None:
        self.session = session
        self.max_consecutive_failures = max_consecutive_failures
        self.consecutive_failures = 0

    def reset(self) -> None:
        self.consecutive_failures = 0

    async def fail(self, action: BaseAction, message: str, exception: Exception | None = None) -> SessionTrajectoryStep:
        _ = await self.session.astep(action)
        obs = await self.session.aobserve()
        return SessionTrajectoryStep(
            action=action,
            obs=obs,
            result=StepResult(success=False, message=message, exception=exception),
        )

    async def execute(self, action: BaseAction) -> SessionTrajectoryStep:
        result = await self.session.astep(action)
        if result.success:
            self.consecutive_failures = 0
        else:
            self.consecutive_failures += 1
            if self.consecutive_failures >= self.max_consecutive_failures:
                # max consecutive failures reached, raise an exception
                if result.exception is None:
                    result.exception = ValueError(result.message)
                raise MaxConsecutiveFailuresError(self.max_consecutive_failures) from result.exception

        obs = await self.session.aobserve()
        return SessionTrajectoryStep(
            action=action,
            obs=obs,
            result=result,
        )



from typing import Unpack

from notte_core.common.notifier import BaseNotifier
from notte_sdk.types import AgentRunRequestDict
from typing_extensions import override

from notte_agent.common.base import BaseAgent
from notte_agent.common.types import AgentResponse


class NotifierAgent(BaseAgent):
    """Agent wrapper that sends notifications after task completion."""

    def __init__(self, agent: BaseAgent, notifier: BaseNotifier):
        super().__init__(session=agent.session)
        self.agent: BaseAgent = agent
        self.notifier: BaseNotifier = notifier

    @override
    async def run(self, **data: Unpack[AgentRunRequestDict]) -> AgentResponse:
        """Run the agent and send notification about the result."""
        result = await self.agent.run(**data)
        self.notifier.notify(data["task"], result)  # pyright: ignore [reportArgumentType]
        return result



import asyncio
from collections.abc import Callable
from enum import StrEnum
from typing import Unpack

from notte_browser.session import NotteSession
from notte_core.agent_types import AgentStepResponse
from notte_core.common.notifier import BaseNotifier
from notte_core.credentials.base import BaseVault
from notte_sdk.types import AgentCreateRequestDict, AgentRunRequest, AgentRunRequestDict

from notte_agent.common.base import BaseAgent
from notte_agent.common.notifier import NotifierAgent
from notte_agent.common.types import AgentResponse
from notte_agent.falco.agent import FalcoAgent
from notte_agent.gufo.agent import GufoAgent


class AgentType(StrEnum):
    FALCO = "falco"
    GUFO = "gufo"


class Agent:
    def __init__(
        self,
        session: NotteSession,
        vault: BaseVault | None = None,
        notifier: BaseNotifier | None = None,
        agent_type: AgentType = AgentType.FALCO,
        **data: Unpack[AgentCreateRequestDict],
    ):
        # just validate the request to create type dependency
        self.data: AgentCreateRequestDict = data
        self.vault: BaseVault | None = vault
        self.notifier: BaseNotifier | None = notifier
        self.session: NotteSession = session
        self.agent_type: AgentType = agent_type

    def create_agent(
        self,
        step_callback: Callable[[AgentStepResponse], None] | None = None,
    ) -> BaseAgent:
        match self.agent_type:
            case AgentType.FALCO:
                agent = FalcoAgent(
                    vault=self.vault,
                    window=self.session.window,
                    storage=self.session.storage,
                    step_callback=step_callback,
                    **self.data,
                )
            case AgentType.GUFO:
                agent = GufoAgent(
                    vault=self.vault,
                    window=self.session.window,
                    # TODO: fix this
                    # step_callback=step_callback,
                    **self.data,
                )
        if self.notifier:
            agent = NotifierAgent(agent, notifier=self.notifier)
        return agent

    async def arun(self, **data: Unpack[AgentRunRequestDict]) -> AgentResponse:
        agent = self.create_agent()
        # validate args
        res = AgentRunRequest.model_validate(data)
        return await agent.run(**res.model_dump())

    def run(self, **data: Unpack[AgentRunRequestDict]) -> AgentResponse:
        return asyncio.run(self.arun(**data))



from notte_browser.session import SessionTrajectoryStep
from notte_core.agent_types import AgentStepResponse
from notte_core.browser.observation import Observation, TrajectoryProgress
from pydantic import BaseModel, Field

from notte_agent.common.types import AgentTrajectoryStep


class AgentTrajectoryHistory(BaseModel):
    max_steps: int
    steps: list[AgentTrajectoryStep] = Field(default_factory=list)

    def reset(self) -> None:
        self.steps = []

    def add_step(self, agent_response: AgentStepResponse, step: SessionTrajectoryStep) -> None:
        step.obs.progress = TrajectoryProgress(
            max_steps=self.max_steps,
            # +1 because we are adding the new step
            current_step=len(self.steps) + 1,
        )
        self.steps.append(
            AgentTrajectoryStep(
                agent_response=agent_response,
                action=step.action,
                obs=step.obs,
                result=step.result,
            )
        )

    def observations(self) -> list[Observation]:
        return [step.obs for step in self.steps]

    def last_obs(self) -> Observation:
        if len(self.steps) == 0:
            raise ValueError("No steps in trajectory")
        return self.steps[-1].obs



import datetime as dt
import traceback
import typing
from collections.abc import Callable

import notte_core
from litellm import AllMessageValues
from loguru import logger
from notte_browser.session import NotteSession
from notte_browser.vault import VaultSecretsScreenshotMask
from notte_core.actions import (
    BaseAction,
    CaptchaSolveAction,
    CompletionAction,
)
from notte_core.agent_types import AgentStepResponse
from notte_core.common.config import NotteConfig, RaiseCondition
from notte_core.common.telemetry import track_usage
from notte_core.common.tracer import LlmUsageDictTracer
from notte_core.credentials.base import BaseVault, LocatorAttributes
from notte_core.llms.engine import LLMEngine
from notte_core.profiling import profiler
from notte_sdk.types import AgentRunRequest, AgentRunRequestDict
from typing_extensions import override

from notte_agent.common.base import BaseAgent
from notte_agent.common.conversation import Conversation
from notte_agent.common.perception import BasePerception
from notte_agent.common.prompt import BasePrompt
from notte_agent.common.safe_executor import SafeActionExecutor
from notte_agent.common.trajectory_history import AgentTrajectoryHistory
from notte_agent.common.types import AgentResponse
from notte_agent.common.validator import CompletionValidator

# #########################################################
# ############### Possible improvements ###################
# #########################################################

# TODO: improve agent memory (e.g. add a memory manager with RAG)
# TODO: use tooling calling for LLM providers that support it
# TODO: file upload/download
# TODO: DIFF rendering module for DOM changes
# TODO: remove base 64 images from current state (reduce token usage)


class NotteAgent(BaseAgent):
    @track_usage("local.agent.create")
    def __init__(
        self,
        prompt: BasePrompt,
        perception: BasePerception,
        config: NotteConfig,
        session: NotteSession,
        vault: BaseVault | None = None,
        step_callback: Callable[[AgentStepResponse], None] | None = None,
    ):
        super().__init__(session=session)
        self.config: NotteConfig = config
        self.llm_tracer: LlmUsageDictTracer = LlmUsageDictTracer()
        self.llm: LLMEngine = LLMEngine(model=self.config.reasoning_model, tracer=self.llm_tracer)
        self.perception: BasePerception = perception
        self.prompt: BasePrompt = prompt
        self.step_callback: Callable[[AgentStepResponse], None] | None = step_callback
        self.step_executor: SafeActionExecutor = SafeActionExecutor(session=self.session)
        # validator a LLM as a Judge that validates the agent's attempt at completing the task (i.e. `CompletionAction`)
        self.validator: CompletionValidator = CompletionValidator(
            llm=self.llm, perception=self.perception, use_vision=self.config.use_vision
        )

        # ####################################
        # ########### Vault Setup ############
        # ####################################

        # vaults are used to safely input credentials into the sessions without leaking them to the LLM (text + screenshots)
        self.vault: BaseVault | None = vault
        if self.vault is not None:
            # hide vault leaked credentials within llm inputs
            self.llm.structured_completion = self.vault.patch_structured_completion(0, self.vault.get_replacement_map)(  # pyright: ignore [reportAttributeAccessIssue]
                self.llm.structured_completion
            )
            # hide vault leaked credentials within screenshots
            self.session.window.screenshot_mask = VaultSecretsScreenshotMask(vault=self.vault)

        # ####################################
        # ######### Conversation Setup #######
        # ####################################

        self.conv: Conversation = Conversation(
            convert_tools_to_assistant=True,
            autosize=True,
            model=self.config.reasoning_model,
        )
        self.trajectory: AgentTrajectoryHistory = AgentTrajectoryHistory(max_steps=self.config.max_steps)
        self.created_at: dt.datetime = dt.datetime.now()

    async def action_with_credentials(self, action: BaseAction) -> BaseAction:
        """Replace credentials in the action if the vault contains credentials"""
        if self.vault is not None and self.vault.contains_credentials(action):
            locator = await self.session.locate(action)
            if locator is not None:
                # compute locator attributes
                attr_type = await locator.get_attribute("type")
                autocomplete = await locator.get_attribute("autocomplete")
                outer_html = await locator.evaluate("el => el.outerHTML")
                attrs = LocatorAttributes(type=attr_type, autocomplete=autocomplete, outerHTML=outer_html)
                # replace credentials
                action = self.vault.replace_credentials(
                    action,
                    attrs,
                    self.session.snapshot,
                )
        return action

    @track_usage("local.agent.reset")
    def reset(self) -> None:
        self.conv.reset()
        self.trajectory.reset()
        self.step_executor.reset()
        self.created_at = dt.datetime.now()

    def output(self, answer: str, success: bool) -> AgentResponse:
        return AgentResponse(
            created_at=self.created_at,
            closed_at=dt.datetime.now(),
            answer=answer,
            success=success,
            trajectory=self.trajectory.steps,
            llm_messages=self.conv.messages(),
            llm_usage=self.llm_tracer.usage,
        )

    @track_usage("local.agent.messages.get")
    async def get_messages(self, task: str) -> list[AllMessageValues]:
        """
        Formats a trajectory into a list of messages for the LLM.

        For every resonning model step, the conversation is reset.
        The conversation follows the following format:

        ### Setup messages
        - [system]        : system prompt containing the initial instructions + action tool calls info
        - [user]          : user request task (e.g. "Find the latest news about AI")
        ### Trajectory messages (one for every step in the trajectory)
            - [assistant] : agent step response (containing memory,state & next action to take)
            - [user]      : session step execution result (success/failure + info message)
        ### DOM perception & final intent message
        - [user]          : DOM perception (browser metadata, page DOM elements, interactive actions, etc.)
        - [user]          : final intent message (e.g. "Select the best action based on whatever I gave you in before")

        /!\\ If `use_vision` is enabled, the DOM perception message will contain a screenshot of the page.
        """
        self.conv.reset()
        system_msg, task_msg = self.prompt.system(), self.prompt.task(task)
        if self.vault is not None:
            system_msg += "\n" + self.vault.instructions()
        self.conv.add_system_message(content=system_msg)
        self.conv.add_user_message(content=task_msg)
        # if no steps in trajectory, add the start trajectory message
        if len(self.trajectory.steps) == 0:
            self.conv.add_user_message(content=self.prompt.empty_trajectory())
            return self.conv.messages()
        # otherwise, add all past trajectorysteps to the conversation
        for step in self.trajectory.steps:
            # TODO: choose if we want this to be an assistant message or a tool message
            # self.conv.add_tool_message(step.agent_response, tool_id="step")
            step_json = step.agent_response.model_dump_json(exclude_none=True)
            self.conv.add_assistant_message(step_json)
            # add step execution status to the conversation
            step_result_content = self.perception.perceive_action_result(
                step.action, step.result, include_ids=False, include_data=True
            )
            self.conv.add_user_message(content=step_result_content)
            # NOTE: if you want to include the full observation (not only structured data), you can do it like this:
            # self.conv.add_user_message(
            #     content=self.perception.perceive(obs),
            #     image=(obs.screenshot if self.config.use_vision else None),
            # )
        last_obs = self.trajectory.last_obs()
        self.conv.add_user_message(
            content=self.perception.perceive(last_obs),
            image=(last_obs.screenshot.bytes() if self.config.use_vision else None),
        )
        self.conv.add_user_message(self.prompt.select_action())
        return self.conv.messages()

    @profiler.profiled()
    @track_usage("local.agent.step")
    async def step(self, request: AgentRunRequest) -> CompletionAction | None:
        """Execute a single step of the agent"""
        messages = await self.get_messages(request.task)
        response: AgentStepResponse = await self.llm.structured_completion(
            messages, response_format=AgentStepResponse, use_strict_response_format=False
        )

        if self.step_callback is not None:
            self.step_callback(response)

        if self.config.verbose:
            logger.trace(f"🔍 LLM response:\n{response}")
        # log the agent state to the terminal
        response.live_log_state()

        # execute the action
        match response.action:
            case CaptchaSolveAction() if not self.session.window.resource.options.solve_captchas:
                # if the session doesnt solve captchas => fail immediately
                error_msg = f"Agent encountered {response.action.captcha_type} captcha but session doesnt solve captchas: create a session with solve_captchas=True"
                return CompletionAction(success=False, answer=error_msg)
            case CompletionAction(success=False, answer=answer):
                # agent failed to complete the task => fail immediately
                logger.error(f"🚨 Agent terminated early with failure: {answer}")
                return CompletionAction(success=False, answer=answer)
            case CompletionAction(success=True, answer=answer):
                # Sucessful execution and need to validate the output
                logger.info(f"🔥 Validating agent output:\n{answer}")
                val_result = await self.validator.validate(
                    output=response.action,
                    history=self.trajectory,
                    task=request.task,
                    response_format=request.response_format,
                )
                if val_result.success:
                    # Successfully validated the output
                    logger.info("✅ Task completed successfully")
                    return response.action
                logger.error(f"🚨 Agent validation failed: {val_result.message}. Continuing...")
                agent_failure_msg = f"""Answer validation failed: {val_result.message}. Continuing...
                CRITICAL: If you think this validation is wrong: argue why the task is finished, or
                perform actions that would prove it is.
                """
                # make sure to add the
                session_step = await self.step_executor.fail(response.action, agent_failure_msg)
            case _:
                # The action is a regular action => execute it (default case)
                action_with_credentials = await self.action_with_credentials(response.action)
                session_step = await self.step_executor.execute(action_with_credentials)
        # Successfully executed the action => add to trajectory
        self.trajectory.add_step(response, session_step)
        step_msg = self.perception.perceive_action_result(response.action, session_step.result, include_ids=True)
        logger.info(f"{step_msg}\n\n")
        return None

    @profiler.profiled()
    @track_usage("local.agent.run")
    @override
    async def run(self, **data: typing.Unpack[AgentRunRequestDict]) -> AgentResponse:
        request = AgentRunRequest.model_validate(data)
        logger.trace(f"Running task: {request.task}")
        self.created_at = dt.datetime.now()
        try:
            return await self._run(request)
        except Exception as e:
            if self.config.raise_condition is RaiseCondition.NEVER:
                return self.output(f"Failed due to {e}: {traceback.format_exc()}", False)
            raise e

    async def _run(self, request: AgentRunRequest) -> AgentResponse:
        """Execute the task with maximum number of steps"""
        # change this to DEV if you want more explicit error messages
        # when you are developing your own agent
        notte_core.set_error_mode("agent")
        if request.url is not None:
            request.task = f"Start on '{request.url}' and {request.task}"

        if self.session.storage is not None:
            request.task = f"{request.task} {self.session.storage.instructions()}"

        for step in range(self.config.max_steps):
            logger.info(f"💡 Step {step}")
            completion_action = await self.step(request)
            if completion_action is not None:
                return self.output(completion_action.answer, completion_action.success)

        error_msg = f"Failed to solve task in {self.config.max_steps} steps"
        logger.info(f"🚨 {error_msg}")
        notte_core.set_error_mode("developer")
        return self.output(error_msg, False)



from notte_core.actions import BaseAction
from notte_core.browser.observation import Observation, StepResult, TrajectoryProgress
from notte_core.browser.snapshot import SnapshotMetadata
from notte_core.data.space import DataSpace
from notte_core.space import ActionSpace
from typing_extensions import override

from notte_agent.common.perception import BasePerception, trim_message


class FalcoPerception(BasePerception):
    @override
    def perceive_metadata(self, metadata: SnapshotMetadata, progress: TrajectoryProgress) -> str:
        return f"""
* Current url: {metadata.url}
* Current page title: {metadata.title}
* Current date and time: {metadata.timestamp.strftime("%Y-%m-%d %H:%M:%S")}
* Available tabs:
{metadata.tabs}
* Current step: {progress.current_step}/{progress.max_steps}
"""

    @override
    def perceive(self, obs: Observation) -> str:
        if obs.progress is None:
            raise ValueError("Observation has no progress")
        px_above = obs.metadata.viewport.pixels_above
        px_below = obs.metadata.viewport.pixels_below

        more_above = f"... {px_above} pixels above - scroll or scrape content to see more ..."
        more_below = f"... {px_below} pixels below - scroll or scrape content to see more ..."
        return f"""
You will see the following only once. If you need to remember it and you dont know it yet, write it down in the memory.

[Relevant metadata]
{self.perceive_metadata(obs.metadata, obs.progress)}

[Interaction elements and context]
[Start of page]
{more_above if px_above > 0 else ""}
{self.perceive_actions(obs.space)}
{more_below if px_below > 0 else ""}
[End of page]
"""

    @override
    def perceive_actions(self, space: ActionSpace) -> str:
        return space.description

    @override
    def perceive_data(self, data: DataSpace | None, only_structured: bool = True) -> str:
        if data is None:
            return ""
        if only_structured:
            structured_data = data.structured
            if structured_data is None or not structured_data.success or structured_data.data is None:
                error_msg = f" with error: {structured_data.error}" if structured_data is not None else ""
                return f"Scraping failed{error_msg}. Please try again with different instructions."
            return f"""
Extracted JSON data:
```json
{structured_data.data.model_dump_json()}
```
"""
        return f"""
Data scraped from current page view:
```markdown
{data.markdown or "No valid data to display"}
```
"""

    @override
    def perceive_action_result(
        self,
        action: BaseAction,
        result: StepResult,
        include_ids: bool = False,
        include_data: bool = True,
    ) -> str:
        id_str = f" with id={action.id}" if include_ids else ""
        if not result.success:
            err_msg = trim_message(result.message)
            return f"❌ action '{action.name()}'{id_str} failed with error: {err_msg}"
        success_msg = f"✅ action '{action.name()}'{id_str} succeeded: '{action.execution_message()}'"
        if include_data:
            return f"{success_msg}{self.perceive_data(result.data, only_structured=True)}"
        return success_msg



import datetime as dt
from typing import Annotated

from litellm import AllMessageValues
from notte_browser.session import SessionTrajectoryStep
from notte_core.agent_types import AgentStepResponse
from notte_core.browser.observation import Screenshot
from notte_core.common.config import ScreenshotType, config
from notte_core.common.tracer import LlmUsageDictTracer
from notte_core.utils.webp_replay import ScreenshotReplay, WebpReplay
from pydantic import BaseModel, Field, computed_field
from typing_extensions import override


class AgentTrajectoryStep(SessionTrajectoryStep):
    agent_response: AgentStepResponse


class AgentResponse(BaseModel):
    success: bool
    answer: str
    trajectory: list[AgentTrajectoryStep]
    # logging information
    created_at: Annotated[dt.datetime, Field(description="The creation time of the agent")]
    closed_at: Annotated[dt.datetime, Field(description="The closed time of the agent")]
    status: str = "closed"
    # only used for debugging purposes
    llm_messages: list[AllMessageValues]
    llm_usage: list[LlmUsageDictTracer.LlmUsage]

    @computed_field
    @property
    def duration_in_s(self) -> float:
        return (self.closed_at - self.created_at).total_seconds()

    @computed_field
    @property
    def steps(self) -> list[AgentStepResponse]:
        return [step.agent_response for step in self.trajectory]

    @override
    def __str__(self) -> str:
        return (
            f"AgentResponse(success={self.success}, duration_in_s={round(self.duration_in_s, 2)}, answer={self.answer})"
        )

    def screenshots(self) -> list[Screenshot]:
        return [step.obs.screenshot for step in self.trajectory]

    def replay(self, step_texts: bool = True, screenshot_type: ScreenshotType = config.screenshot_type) -> WebpReplay:
        screenshots: list[bytes] = []
        texts: list[str] = []

        for step in self.trajectory:
            screenshots.append(step.obs.screenshot.bytes(screenshot_type))
            texts.append(step.agent_response.state.next_goal)

        if len(screenshots) == 0:
            raise ValueError("No screenshots found in agent trajectory")

        if step_texts:
            return ScreenshotReplay.from_bytes(screenshots).get(step_text=texts)  # pyright: ignore [reportArgumentType]
        else:
            return ScreenshotReplay.from_bytes(screenshots).get()

    @override
    def __repr__(self) -> str:
        return self.__str__()



from pathlib import Path

from notte_agent.falco.prompt import FalcoPrompt


class GufoPrompt(FalcoPrompt):
    def __init__(self):
        super().__init__(prompt_file=Path(__file__).parent / "system.md")



import typing
from collections.abc import Callable

from notte_browser.session import NotteSession
from notte_browser.window import BrowserWindow
from notte_core.agent_types import AgentStepResponse
from notte_core.common.config import NotteConfig
from notte_core.credentials.base import BaseVault
from notte_sdk.types import AgentCreateRequest, AgentCreateRequestDict
from pydantic import field_validator

from notte_agent.agent import NotteAgent
from notte_agent.gufo.perception import GufoPerception
from notte_agent.gufo.prompt import GufoPrompt


class GufoConfig(NotteConfig):
    enable_perception: bool = True

    @classmethod
    @field_validator("enable_perception")
    def check_perception(cls, value: bool) -> bool:
        if not value:
            raise ValueError("Perception should be enabled for gufo. Don't set this argument to `False`.")
        return value


class GufoAgent(NotteAgent):
    def __init__(
        self,
        window: BrowserWindow,
        vault: BaseVault | None = None,
        step_callback: Callable[[AgentStepResponse], None] | None = None,
        **data: typing.Unpack[AgentCreateRequestDict],
    ):
        _ = AgentCreateRequest.model_validate(data)
        config: GufoConfig = GufoConfig.from_toml(**data)
        session = NotteSession(window=window, enable_perception=True)
        super().__init__(
            prompt=GufoPrompt(),
            perception=GufoPerception(),
            config=config,
            session=session,
            vault=vault,
            step_callback=step_callback,
        )
